// -*- tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set ts=8 sw=4 et sts=4:
#ifndef DUNE_ELASTICITY_MATERIALS_GEOMETRICALLY_EXACT_ST_VENANT_KIRCHHOFF_MATERIAL_HH
#define DUNE_ELASTICITY_MATERIALS_GEOMETRICALLY_EXACT_ST_VENANT_KIRCHHOFF_MATERIAL_HH

#include <dune/fufem/quadraturerules/quadraturerulecache.hh>
#include <dune/fufem/symmetrictensor.hh>
#include <dune/fufem/functions/virtualgridfunction.hh>
#include <dune/fufem/assemblers/localassemblers/geomexactstvenantkirchhofffunctionalassembler.hh>
#include <dune/fufem/assemblers/localassemblers/geomexactstvenantkirchhoffoperatorassembler.hh>

#include <dune/elasticity/materials/material.hh>
#include <dune/elasticity/common/elasticityhelpers.hh>

/* \brief Class representing a hyperelastic geometrically exact St.Venant Kirchhoff material.
 *
 * \tparam Basis    Global basis that is used for the spatial discretization. 
 *                  (This is not nice but I need a LocalFiniteElement for the local Hessian assembler :-( )
*/
template <class Basis>
class GeomExactStVenantMaterial : public Material<Basis>
{
    typedef Material<Basis> Base;
public:
    typedef typename Basis::GridView::Grid GridType;
    typedef typename Base::GlobalBasis GlobalBasis;
    typedef typename Base::Lfe Lfe;
    typedef typename Base::LocalLinearization LocalLinearization;
    typedef typename Base::LocalHessian LocalHessian;
    typedef typename Base::VectorType VectorType;
    typedef typename Base::GridFunction GridFunction;
    typedef typename Base::ReturnType ReturnType;


private:
    using Base::dim;
    typedef typename GridType::ctype ctype;
    typedef GeomExactStVenantKirchhoffFunctionalAssembler<GridType,Lfe> GeomLinearization;
    typedef GeomExactStVenantKirchhoffOperatorAssembler<GridType, Lfe, Lfe> GeomHessian;
    typedef typename GridType::template Codim<0>::Geometry::LocalCoordinate LocalCoordinate; 
    typedef typename GridType::template Codim<0>::LeafIterator ElementIterator; 

public:
    GeomExactStVenantMaterial() :
        E_(1.0),
        nu_(0.3)
    {}

    GeomExactStVenantMaterial(const Basis& basis, ReturnType E, ReturnType nu) :
        Base(basis),
        E_(E),
        nu_(nu)
    {
        localLinearization_ = Dune::shared_ptr<GeomLinearization>(new GeomLinearization(E,nu));
        localHessian_ = Dune::shared_ptr<GeomHessian>(new GeomHessian(E,nu));
    }

    void setup(ReturnType E, ReturnType nu, const Basis& basis)
    {
        E_ = E; 
        nu_ = nu;

        localLinearization_ = Dune::shared_ptr<GeomLinearization>(new GeomLinearization(E,nu));
        localHessian_ = Dune::shared_ptr<GeomHessian>(new GeomHessian(E,nu));

        this->basis_ = &basis;
    }

    //! Evaluate the strain energy
    ReturnType energy(const GridFunction* displace)
    {
        ReturnType energy=0;
        const GridType& grid = this->basis_->getGridView().grid();

        ElementIterator eIt = grid.template leafbegin<0>();
        ElementIterator eItEnd = grid.template leafend<0>();

        for (;eIt != eItEnd; ++eIt) {

            // get quadrature rule
            QuadratureRuleKey quad1(this->basis_->getLocalFiniteElement(*eIt));
            QuadratureRuleKey quadKey = quad1.derivative().square().square();

            const Dune::template QuadratureRule<ctype, dim>& quad = QuadratureRuleCache<ctype, dim>::rule(quadKey);

            // loop over quadrature points
            for (size_t pt=0; pt < quad.size(); ++pt) {

                // get quadrature point
                const LocalCoordinate& quadPos = quad[pt].position();

                // get integration factor
                const ctype integrationElement = eIt->geometry().integrationElement(quadPos);

                // evaluate displacement gradient at the quadrature point
                typename GridFunction::DerivativeType localDispGrad;

                if (displace->isDefinedOn(*eIt))
                    displace->evaluateDerivativeLocal(*eIt, quadPos, localDispGrad);
                else
                    displace->evaluateDerivative(eIt->geometry().global(quadPos),localDispGrad);

                // Compute the nonlinear strain tensor from the deformation gradient
                SymmetricTensor<dim,ReturnType> strain;
                Dune::Elasticity::computeNonlinearStrain(localDispGrad,strain);

                // and the stress
                SymmetricTensor<dim,ReturnType> stress = hookeTimesStrain(strain);

                ctype z = quad[pt].weight()*integrationElement;
                energy += (stress*strain)*z;
            }
        }
            
        return 0.5*energy;
    }

    //! Return the local assembler of the first derivative of the strain energy 
    LocalLinearization& firstDerivative(GridFunction* displace)
    {
        localLinearization_->setConfiguration(displace);
        return *localLinearization_;
    }

    //! Return the local assembler of the second derivative of the strain energy 
    LocalHessian& secondDerivative(GridFunction* displace)
    {
        localHessian_->setConfiguration(displace);
        return *localHessian_;
    }

private:
    //! First derivative of the strain energy
    Dune::shared_ptr<GeomLinearization> localLinearization_;

    //! Second derivative of the strain energy
    Dune::shared_ptr<GeomHessian> localHessian_;

    //! Elasticity modulus
    ReturnType E_;
    //! Poisson ratio
    ReturnType nu_;
    
    //! Compute linear elastic stress from the strain
    SymmetricTensor<dim,ReturnType> hookeTimesStrain(const SymmetricTensor<dim,ReturnType>& strain) const {

        SymmetricTensor<dim,ReturnType> stress = strain;
        stress *= E_/(1+nu_);

        ReturnType f = E_*nu_ / ((1+nu_)*(1-2*nu_)) * strain.trace();

        for (int i=0; i<dim; i++)
            stress(i,i) += f;

        return stress;
    }
};

#endif
